/*
 * Copyright (c) 1999-2003 WIDE Project
 * All rights reserved.
 *
 * Author : Akimichi OGAWA (akimichi@sfc.wide.ad.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code MUST retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form MUST reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    MUST display the following acknowledgement:
 *      This product includes software developed by Akimichi OGAWA.
 * 4. The name of the author MAY NOT be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
*/

#ifdef HAVE_CONFIG_H
#include <dvts-config.h>
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/ioctl.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>

#include "hdvsend_param.h"
#include "multicast.h"

static void _IPv4_multicast_set_ttl (int soc, int val);
static void _IPv4_multicast_set_if (int soc, char *ifname);
static void _IPv6_multicast_set_ttl (int soc, int val);
static void _IPv6_multicast_set_if (int soc, char *ifname);

void
multicast_settings (struct hdvsend_param *hdvsend_param)
{
  struct dest_obj *dest;
  struct sockaddr *s_addr;

  assert(hdvsend_param);

  for (dest = hdvsend_param->dest_list.tqh_first; dest != NULL;
       dest = (struct dest_obj *)dest->chain.tqe_next){

    s_addr = (struct sockaddr *)&dest->rtp_s_addr;
    switch (s_addr->sa_family) {
      case AF_INET:
        if (IN_MULTICAST(ntohl(((struct sockaddr_in *)s_addr)->sin_addr.s_addr))) {
          _IPv4_multicast_set_ttl(dest->soc, dest->multi_ttl);
          _IPv4_multicast_set_if(dest->soc, dest->multi_if);
        }
        break;

      case AF_INET6:
        if (IN6_IS_ADDR_MULTICAST(&((struct sockaddr_in6 *)s_addr)->sin6_addr)) {
          _IPv6_multicast_set_ttl(dest->soc, dest->multi_ttl);
          _IPv6_multicast_set_if(dest->soc, dest->multi_if);
        }
        break;

      default:
        printf("multicast_settings : UNKNOWN sa_family\n");
        assert(0);
        break;
    }
  }
}

static void
_IPv4_multicast_set_ttl (int soc, int val){
  u_char ttl = val;
  if (setsockopt(soc, IPPROTO_IP, IP_MULTICAST_TTL,
                 (void *)&ttl, sizeof(ttl)) < 0) {
    perror("setsockopt multicast ttl");
  }
}

static void
_IPv6_multicast_set_ttl (int soc, int val){
  int hoplimit = val;
  if (setsockopt(soc, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
                 (void *)&hoplimit, sizeof(hoplimit)) < 0) {
    perror("setsockopt IPv6 multicast hops");
  }
}

static void
_IPv4_multicast_set_if (int soc, char *multi_if){
  struct in_addr addr;

  memset(&addr, 0, sizeof(struct in_addr));

  if (inet_pton(AF_INET, multi_if, &addr) <= 0)  {
    struct ifreq ifreq;

    strncpy(ifreq.ifr_name, multi_if, IFNAMSIZ);
    if (ioctl(soc, SIOCGIFADDR, &ifreq) < 0){
      perror("ioctl SIOCGIOFADDR");
    }
    memcpy(&addr, &((struct sockaddr_in *) &ifreq.ifr_addr)->sin_addr,
	   sizeof(struct in_addr));
  }
  if (setsockopt(soc, IPPROTO_IP, IP_MULTICAST_IF, &addr, sizeof(addr)) < 0) {
    perror("setsockopt IPv4 multicast interface");
  } else {
    char tmp[NI_MAXHOST + 1];
    inet_ntop(AF_INET, (void *)&addr, tmp, NI_MAXHOST);
    printf("set IPv4 multicast interface: %s\n", tmp);
  }
}

static void
_IPv6_multicast_set_if (int soc, char *multi_if){
  int ifindex = if_nametoindex(multi_if);

  if (setsockopt(soc, IPPROTO_IPV6, IPV6_MULTICAST_IF, &ifindex, sizeof(ifindex)) < 0) {
    perror("setsockopt IPv6 multicast interface");
  } else {
    printf("set IPv6 multicast interface: %s\n", multi_if);
  }
}
