/*
 * Copyright (c) 1999-2003 WIDE Project
 * All rights reserved.
 *
 * Author : Akimichi OGAWA (akimichi@sfc.wide.ad.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code MUST retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form MUST reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    MUST display the following acknowledgement:
 *      This product includes software developed by Akimichi OGAWA.
 * 4. The name of the author MAY NOT be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
*/

#ifdef HAVE_CONFIG_H
#include <dvts-config.h>
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>

#include <rtpvar.h>

#include "udp.h"
#include "rtcp.h"
#include "flags.h"

#define MAX_RTCP_PKT_LEN	1500
#define RTCP_SR_INTERVAL        3
#define NTP_BASE		2208988800U

static int _process_rtcp_rr __P((struct hdvsend_param *, char *, int, struct sockaddr *, int));
static int _process_rtcp_bye __P((struct hdvsend_param *, char *, int, struct sockaddr *, int));

int
send_rtcp_sr (struct hdvsend_param *hdvsend_param)
{
  int n;
  char buf[MAX_RTCP_PKT_LEN];
  int buflen;
  struct timeval tv;
  u_int32_t the_ntp_sec = NTP_BASE;

  rtcp_common_t *rtcp_hdr;
  struct sr {
	u_int32_t ssrc;
	u_int32_t ntp_sec;
	u_int32_t ntp_frac;
	u_int32_t rtp_ts;
	u_int32_t psent;
	u_int32_t osent;
	rtcp_rr_t rr[1];
  } *rtcp_sr;

  memset(&buf, 0, sizeof(buf));
  rtcp_hdr = (rtcp_common_t *)buf;
  rtcp_sr  = (struct sr *)&buf[sizeof(rtcp_common_t)];

  buflen = sizeof(rtcp_common_t) + sizeof(struct sr);

  rtcp_hdr->version = RTP_VERSION;
  rtcp_hdr->count = 1;
  rtcp_hdr->pt = RTCP_SR;
  rtcp_sr->ssrc = hdvsend_param->rtp_ssrc;
  gettimeofday(&tv, NULL);
  the_ntp_sec += tv.tv_sec;
  rtcp_sr->ntp_sec = htonl(the_ntp_sec); /* diff 1970/1/1,1900/1/1 */
  rtcp_sr->ntp_frac = htonl(tv.tv_usec);
  rtcp_sr->rtp_ts = htonl(the_ntp_sec/90000);

  n = send_rtcp_pkt(hdvsend_param, (u_int32_t *)buf, buflen);
  if (n < 1) {
    perror("rtcp sendto");
  }

  return(1);
}

int
send_rtcp_bye (struct hdvsend_param *hdvsend_param)
{
  int n;
  rtcp_t rtcp_bye;
  memset(&rtcp_bye, 0, sizeof(rtcp_t));

  rtcp_bye.common.version = RTP_VERSION;
  rtcp_bye.common.count = 1;
  rtcp_bye.common.pt = RTCP_BYE;
  rtcp_bye.r.bye.src[0] = hdvsend_param->rtp_ssrc;

  n = send_rtcp_pkt(hdvsend_param, (u_int32_t *)&rtcp_bye, sizeof(rtcp_t));
  if (n < 1) {
    perror("rtcp sendto");
  }

  return(n);
}

int
rtcp_check(struct hdvsend_param *hdvsend_param, u_int32_t *read_count,
		           struct timeval *tv_prev)
{
	struct timeval tv;
	(*read_count)++;
	/* just do RTCP process sometimes */
	if ((hdvsend_param->flags & USE_RTCP) && *read_count % 1000) {
		if (gettimeofday(&tv, NULL) < 0) {
			perror("gettimeofday failed in rtcp_check");
		} else if ((tv.tv_sec - tv_prev->tv_sec) > RTCP_SR_INTERVAL) {
			send_rtcp_sr(hdvsend_param);
			memcpy(tv_prev, &tv, sizeof(*tv_prev));
		}
		/* process rtcp input */
		process_rtcp(hdvsend_param);
	}
	return(*read_count);
}


/* This sees if there is a RTCP packet in the queue.      */
/* If there is no RTCP packet in the queue, this function */
/* simply returns.                                        */
int
process_rtcp (struct hdvsend_param *hdvsend_param)
{
  char buf[MAX_RTCP_PKT_LEN];
  rtcp_common_t *rtcp_hdr;
  int n;
  struct sockaddr_storage from;
  socklen_t length = sizeof(from);

  struct rtcp_recv_obj *rtcp_recv = NULL;

  if (!(hdvsend_param->flags & USE_RTCP)) {
    return(1);
  }

  memset(buf, 0, sizeof(buf));
  rtcp_hdr = (rtcp_common_t *)buf;

  for (rtcp_recv = hdvsend_param->rtcp_recv_list.tqh_first; rtcp_recv != NULL; 
       rtcp_recv = (struct rtcp_recv_obj *)rtcp_recv->chain.tqe_next) {
    if ((n = recvfrom(rtcp_recv->soc,
                      buf, sizeof(buf),
                      0,
                      (struct sockaddr *)&from, &length)) < 0) {
      if (errno != EAGAIN) {
        perror("rtcp recvfrom");
        return(-1);
      }
      else {
        /* no RTCP packets are in the queue */
        continue;
      }
    }

    switch (rtcp_hdr->pt) {
      case RTCP_RR:
        _process_rtcp_rr(hdvsend_param,
                         buf, n, (struct sockaddr *)&from, length);
        break;

      case RTCP_BYE:
        _process_rtcp_bye(hdvsend_param,
                          buf, n, (struct sockaddr *)&from, length);
        break;

      default:
        break;
    }
  }

  return(1);
}

static int
_process_rtcp_rr (struct hdvsend_param *hdvsend_param,
                  char *buf, int n,
                  struct sockaddr *addr, int length)
{
  rtcp_t *rtcp_rr;
  u_long frac_lost;
  int  pkt_loss_sum;
  float frac_lost_100;
  char s_addr_str[64];
#ifdef	ENABLE_INET6
  int err;
#endif	/* ENABLE_INET6 */

  memset(s_addr_str, 0, sizeof(s_addr_str));

  /* if showing number of packet loss is required */
  if (hdvsend_param->flags & SHOW_PACKETLOSS) {
    rtcp_rr = (rtcp_t *)buf;

    frac_lost = rtcp_rr->r.rr.rr[0].fraction;
    pkt_loss_sum = ntohl(rtcp_rr->r.rr.rr[0].lost << 8);

    /*                                    */
    /* changes 1/255 scale to 1/100 scale */
    /*                                    */
    if (frac_lost != 0) {
      frac_lost_100 = (double)(frac_lost) * 100.0 / 255.0;
    }
    else {
      frac_lost_100 = 0;
    }

#ifdef	ENABLE_INET6
#ifndef	NO_SS_LEN
    err = getnameinfo(addr, addr->sa_len,
#else	/* NO_SS_LEN */
    err = getnameinfo(addr,
                      (addr->sa_family==AF_INET6)?sizeof(struct sockaddr_in6):sizeof(struct sockaddr_in),
#endif	/* NO_SS_LEN */
                      s_addr_str, sizeof(s_addr_str),
                      NULL, 0, NI_NUMERICHOST);
#else	/* ENABLE_INET6 */
    memcpy(s_addr_str,
           (char *)inet_ntoa(((struct sockaddr_in *)addr)->sin_addr),
           sizeof(struct in_addr));
#endif	/* ENABLE_INET6 */
                      
    printf("receiver [%s], pkt lost [%d], pkt loss (%%) [%.2f %%]\n",
           s_addr_str,
           pkt_loss_sum,
           frac_lost_100);
  }

  return(1);
}

static int
_process_rtcp_bye (struct hdvsend_param *hdvsend_param,
                   char *buf, int n,
                   struct sockaddr *addr, int length)
{
  printf("RECEIVED RTCP BYE\n");

  return(1);
}
