/*
 * Copyright (c) 1999-2003 WIDE Project
 * All rights reserved.
 *
 * Author : Akimichi OGAWA (akimichi@sfc.wide.ad.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code MUST retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form MUST reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    MUST display the following acknowledgement:
 *      This product includes software developed by Akimichi OGAWA.
 * 4. The name of the author MAY NOT be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
*/

#ifdef HAVE_CONFIG_H
#include <dvts-config.h>
#endif	/* HAVE_CONFIG */

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <dvtsutil.h>

#include "hdvsend_param.h"
#include "flags.h"
#include "udp.h"
#include "rtpvar.h"
//#include "hoge.h"

//#include "ff_tx.h"

#ifndef DVTS_FF
  #define DVTS_FF
  #include <ff_tx.h>
  //extern  int ff_verbosity;
#endif



static int _prepare_rtp_send  __P((struct hdvsend_param *));
static int _prepare_rtcp_send __P((struct hdvsend_param *));
// static int _prepare_rtcp_recv __P((struct hdvsend_param *));

static int _prepare_rtp_audio_send  __P((struct hdvsend_param *));
static int _prepare_rtcp_audio_send __P((struct hdvsend_param *));
// static int _prepare_rtcp_audio_recv __P((struct hdvsend_param *));


static int _prepare_send_sockaddr __P((char *, char *, struct sockaddr_storage *));

static int _parse_dest_str __P((char *, char *, int, char *, int, int *, char *, int));

static int _add_destination_common __P((struct hdvsend_param *, char *,
			                struct dest_list_head *));

/****** static functions START ******/

static inline u_int8_t 
get_ss_len(const struct sockaddr_storage *ss) 
{
#if defined(HAVE_SS_LEN) || defined(BSD)
  return ss->ss_len;
#else /* HAVE_SS_LEN */
  switch (ss->ss_family) {
  case AF_INET:
    return sizeof(struct sockaddr_in);
#ifdef ENABLE_INET6
  case AF_INET6:
    return sizeof(struct sockaddr_in6);
#endif /* ENABLE_INET6 */
  }
  return 0;
#endif /* HAVE_SS_LEN */
}

static inline int
set_ss_len(struct sockaddr_storage *ss, u_int8_t len) 
{
#if defined(HAVE_SS_LEN) || defined(BSD)
  ss->ss_len = len;
#endif /* HAVE_SS_LEN */
  return (1);
}

static inline int 
init_ss_len(struct sockaddr_storage *ss) 
{
  switch (ss->ss_family) {
  case AF_INET:
    return set_ss_len(ss, sizeof(struct sockaddr_in));
#ifdef ENABLE_INET6
  case AF_INET6:
    return set_ss_len(ss, sizeof(struct sockaddr_in6));
    break;
#endif /* ENABLE_INET6 */
  }
  return (-1);
}

static inline u_int16_t 
get_ss_port(const struct sockaddr_storage *ss)
{
  switch (ss->ss_family) {
  case AF_INET:
    return ntohs(((struct sockaddr_in *)ss)->sin_port);
#ifdef ENABLE_INET6
  case AF_INET6:
    return ntohs(((struct sockaddr_in6 *)ss)->sin6_port);
#endif /* ENABLE_INET6 */
  }
  return (0);
}

static inline int 
set_ss_port(struct sockaddr_storage *ss, u_int16_t port)
{
  switch (ss->ss_family) {
  case AF_INET:
    ((struct sockaddr_in *)ss)->sin_port = htons(port);
    break;
#ifdef ENABLE_INET6
  case AF_INET6:
    ((struct sockaddr_in6 *)ss)->sin6_port = htons(port);
    break;
#endif /* ENABLE_INET6 */
  default:
    return (-1);
  }
  return (1);
}

static inline int 
set_ss_addr_any(struct sockaddr_storage *ss)
{
  switch (ss->ss_family) {
  case AF_INET:
    ((struct sockaddr_in *)ss)->sin_addr.s_addr = htonl(INADDR_ANY);
    break;
#ifdef ENABLE_INET6
  case AF_INET6:
    ((struct sockaddr_in6 *)ss)->sin6_addr = in6addr_any;
    break;
#endif /* ENABLE_INET6 */
  default:
    return (-1);
  }
  return (1);
}

static int
_parse_dest_str(char *dest_str,
                char *dest_ret, int dest_str_len,
                char *serv_ret, int serv_len,
                int *multi_ttl_ret,
                char *multi_ifname_ret, int multi_ifname_len)
{
  int number_of_slash = 0;
  char cp_dest_str[BUFSIZ];
  char *sep_ptr;
  enum {HOST, SERV, TTL, IFNAME};
  char *ret_ptr[4] = {dest_ret, serv_ret, (char *)multi_ttl_ret, multi_ifname_ret};
  int ret_len[4] = {dest_str_len, serv_len, 0, multi_ifname_len};

  char *ptr;

  if (dest_str == NULL){
    return(-1);
  }

  strncpy(cp_dest_str, dest_str, sizeof(cp_dest_str));

  sep_ptr = cp_dest_str;
  for(ptr = strsep(&sep_ptr, "/"); ptr != NULL; ptr = strsep(&sep_ptr, "/")){
    if(ret_ptr[number_of_slash] == NULL){
      return(-1);
    }
    if(number_of_slash == TTL){
      *(int *)ret_ptr[number_of_slash] = atoi(ptr);
    } else {
      strncpy(ret_ptr[number_of_slash], ptr, ret_len[number_of_slash]);
      ret_ptr[number_of_slash][ret_len[number_of_slash]] = '\0';
    }
    number_of_slash++;
  }

  return(1);
}

static int
_add_destination_common(struct hdvsend_param *hdvsend_param, char *buf, struct dest_list_head *head)
{
  struct dest_obj *dest = (struct dest_obj *)malloc(sizeof(struct dest_obj));

  if (dest == NULL) return (-1);

  _parse_dest_str(buf,
                  dest->dest_str,
                  sizeof(dest->dest_str),
                  dest->serv, sizeof(dest->serv),
                  &dest->multi_ttl,
                  dest->multi_if, sizeof(dest->multi_if));

  if ((dest->serv == NULL) || (strlen(dest->serv) <= 0)) {
    strncpy(dest->serv, hdvsend_param->default_serv,
	    sizeof(dest->serv));
  }
  if (dest->multi_ttl < 1) {
    dest->multi_ttl = hdvsend_param->default_multi_ttl;
  }

  if ((dest->multi_if == NULL) || (strlen(dest->multi_if) <= 0)) {
    strncpy(dest->multi_if, hdvsend_param->default_multi_src_addr,
	    sizeof(dest->multi_if));
  }

  if (hdvsend_param->flags & USE_IPV4_ONLY) {
    dest->rtp_s_addr.ss_family = AF_INET;
    dest->rtcp_s_addr.ss_family = AF_INET;
    set_ss_len(&dest->rtp_s_addr, sizeof(struct sockaddr_in));
    set_ss_len(&dest->rtcp_s_addr, sizeof(struct sockaddr_in));
  }
  else if (hdvsend_param->flags & USE_IPV6_ONLY) {
    dest->rtp_s_addr.ss_family = AF_INET6;
    dest->rtcp_s_addr.ss_family = AF_INET6;
    set_ss_len(&dest->rtp_s_addr, sizeof(struct sockaddr_in6));
    set_ss_len(&dest->rtcp_s_addr, sizeof(struct sockaddr_in6));
  }
  else {
    dest->rtp_s_addr.ss_family = PF_UNSPEC;
    dest->rtcp_s_addr.ss_family = PF_UNSPEC;
  }

  TAILQ_INSERT_TAIL(head, dest, chain);

  return(1);
}

static int
_send_rtp_common(struct dest_list_head *dest_list, u_int32_t *buf,
		 int buflen)
{
  int n;
  struct dest_obj *dest;
  //testprint();

  #ifdef FF_DVTS   
  //put interface to FECFRAME
  n = fecFrameSenderDVTS(dest_list, buf, buflen);
  #endif


  #ifndef FF_DVTS
  for (dest = dest_list->tqh_first; dest != NULL;
       dest = (struct dest_obj *)dest->chain.tqe_next){

    n = sendto(dest->soc, buf, buflen, 0,
	       (struct sockaddr *)&dest->rtp_s_addr,
	       get_ss_len(&dest->rtp_s_addr));

    if (n < 1) {
      perror("send_pkt : sendto");
    }

    dest->pkt_count++;
    dest->oct_count += n / sizeof(u_int32_t);
  }
  #endif

  return (1);
}

static int
_send_rtcp_common(struct dest_list_head *dest_list, u_int32_t *buf,
		  int buflen)
{
  int n;
  struct dest_obj *dest = NULL;
  rtcp_t *rtcpbuf = (rtcp_t *)buf;

  for (dest = dest_list->tqh_first; dest != NULL;
       dest = (struct dest_obj *)dest->chain.tqe_next){

    if(rtcpbuf->common.pt == RTCP_SR){
      rtcpbuf->r.sr.psent = ntohl(dest->pkt_count);
      rtcpbuf->r.sr.osent = ntohl(dest->oct_count);
    }

    n = sendto(dest->soc, buf, buflen, 0,
	       (struct sockaddr *)&dest->rtcp_s_addr,
	       get_ss_len(&dest->rtcp_s_addr));

    if (n < 1) {
      perror("send_rtcp : sendto");
    }
  }

  return(1);
}

/****** static functions END ******/

int
add_destination(struct hdvsend_param *hdvsend_param, char *buf)
{
  return(_add_destination_common(hdvsend_param, buf,
			         &hdvsend_param->dest_list));
}

int
add_audio_only_destination(struct hdvsend_param *hdvsend_param, char *buf)
{
  return(_add_destination_common(hdvsend_param, buf,
			         &hdvsend_param->dest_audio_list));
}

int
send_pkt(struct hdvsend_param *hdvsend_param, u_int32_t *buf, int buflen)
{
  return(_send_rtp_common(&hdvsend_param->dest_list, buf, buflen));
}

int
send_audio_pkt(struct hdvsend_param *hdvsend_param, u_int32_t *buf, int buflen)
{
  return(_send_rtp_common(&hdvsend_param->dest_audio_list, buf, buflen));
}

int
send_rtcp_pkt(struct hdvsend_param *hdvsend_param, u_int32_t *buf, int buflen)
{
  return(_send_rtcp_common(&hdvsend_param->dest_list, buf, buflen));
}

int
send_rtcp_audio_pkt(struct hdvsend_param *hdvsend_param,
                    u_int32_t *buf,
                    int buflen)
{
  return(_send_rtcp_common(&hdvsend_param->dest_audio_list, buf, buflen));
}

int
prepare_udp_socket (struct hdvsend_param *hdvsend_param)
{
  if (_prepare_rtp_send(hdvsend_param) < 0) {
    printf("failed to create RTP send socket\n");
    return(-1);
  }

  if (hdvsend_param->flags & SEPARATED_AUDIO_STREAM) {
    if (_prepare_rtp_audio_send(hdvsend_param) < 0) {
      printf("failed to create RTP audio send socket\n");
      return(-1);
    }
  }

  if (hdvsend_param->flags & USE_RTCP) {
    if (_prepare_rtcp_send(hdvsend_param) < 0) {
      printf("failed to create RTCP send socket\n");
      return(-1);
    }
#if 0
    if (_prepare_rtcp_recv(hdvsend_param) < 0) {
      printf("failed to create RTCP send socket\n");
      return(-1);
    }
#endif
    if (hdvsend_param->flags & SEPARATED_AUDIO_STREAM) {
      if (_prepare_rtcp_audio_send(hdvsend_param) < 0) {
        printf("failed to create RTCP audio socket\n");
        return(-1);
      }
#if 0
      if (_prepare_rtcp_audio_recv(hdvsend_param) < 0) {
        printf("failed to create RTCP audio receive socket\n");
        return(-1);
      }
#endif
    }
  }

  return(1);
}

static int
_prepare_rtp_send_common (struct dest_list_head *dest_list_head, char *default_serv)
{
  int ret;
  struct dest_obj *dest;

  for (dest = dest_list_head->tqh_first; dest != NULL;
       dest = (struct dest_obj *)dest->chain.tqe_next){

    if (strlen(dest->serv) == 0) strncpy(dest->serv, default_serv, NI_MAXSERV);
    ret = _prepare_send_sockaddr(dest->dest_str, dest->serv,
                                 &dest->rtp_s_addr);
    if (ret < 0) {
      return(ret);
    }

    printf("RTP  dest/serv : %s/%s\n", dest->dest_str, dest->serv);

    dest->soc = socket(dest->rtp_s_addr.ss_family, SOCK_DGRAM, 0);

    if (dest->soc < 0) {
      perror("socket");
      return(dest->soc);
    }

#ifdef _SET_TOS_1
    if (dest->rtp_s_addr.ss_family == AF_INET) {
      unsigned char tos = 0x10;
      if (setsockopt(dest->soc, IPPROTO_IP, IP_TOS, (void *)&tos, sizeof(tos)) < 0) {
	perror("Set IP TOS");
      } else {
	printf("Set IP TOS to 1\n");
      }
    } else if (dest->rtp_s_addr.ss_family == AF_INET6) {
      int on = 1;
      
      struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)&dest->rtp_s_addr;
      sin6->sin6_flowinfo = htonl(0x01000000);
      #ifdef LINUX
      if (setsockopt(dest->soc, SOL_IPV6, 33, (void *)&on, sizeof(on)) < 0) {
	perror("Set IPv6 Flowinfo send");
      } else {
	printf("Set IPV6 Flowinfo send\n");
      }
      #endif
    }
#endif
  }
  return(1);
}
static int
_prepare_rtp_send (struct hdvsend_param *hdvsend_param)
{
  printf("DESTINATIONS with AUDIO & VIDEO\n");
  return(_prepare_rtp_send_common(&hdvsend_param->dest_list, hdvsend_param->default_serv));
}

static int
_prepare_rtp_audio_send (struct hdvsend_param *hdvsend_param)
{
  printf("DESTINATIONS with AUDIO ONLY\n");
  return(_prepare_rtp_send_common(&hdvsend_param->dest_audio_list, hdvsend_param->default_audio_serv));
}

/********** RTCP ************/

static int
_add_rtcp_recv (struct rtcp_recv_list_head *rtcp_recv_list, const struct sockaddr_storage *saddr)
{
  struct rtcp_recv_obj *rtcp_recv = NULL;
  int _found = 0;
  int af = saddr->ss_family;
  u_int16_t port = get_ss_port(saddr);

  for (rtcp_recv = rtcp_recv_list->tqh_first; rtcp_recv != NULL; rtcp_recv = rtcp_recv->chain.tqe_next) {
    if (rtcp_recv->s_addr.ss_family == af
	&& get_ss_port(&rtcp_recv->s_addr) == port) {
      _found = 1;
      break;
    }
  }
  
  if (_found == 0) {
    int on = 1;

    rtcp_recv = (struct rtcp_recv_obj *)malloc(sizeof(struct rtcp_recv_obj));    
    if (rtcp_recv == NULL) {
      perror("add_rtcp_recv: malloc");
      return (-1);
    }

    memset(&rtcp_recv->s_addr, 0, sizeof(struct sockaddr_storage));
    rtcp_recv->s_addr.ss_family = saddr->ss_family;
    init_ss_len(&rtcp_recv->s_addr);
    set_ss_port(&rtcp_recv->s_addr, get_ss_port(saddr));

    if (set_ss_addr_any(&rtcp_recv->s_addr) < 0) {
      fprintf(stderr, "add_rtcp_recv: unknown address family");
      free(rtcp_recv);
      return -1;
    }
    
    rtcp_recv->soc = socket(rtcp_recv->s_addr.ss_family, SOCK_DGRAM, IPPROTO_UDP);
    if (rtcp_recv->soc == -1) {
      perror("add_rtcp_recv: socket");
      free(rtcp_recv);
      return (-1);
    }

#ifdef ENABLE_INET6
    if (rtcp_recv->s_addr.ss_family == AF_INET6) {
#ifdef IPV6_V6ONLY
	if (setsockopt(rtcp_recv->soc, IPPROTO_IPV6, IPV6_V6ONLY, &on, sizeof(on)) < 0) {
	  perror("add_rtcp_recv: setsockopt IPV6_V6ONLY");
	  free (rtcp_recv);
	  return -1;
	}
#endif /* IPv6_V6ONLY */      
    }
#endif /* ENABLE_INET6 */

    if (setsockopt(rtcp_recv->soc, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) {
      perror("add_rtcp_recv: setsockopt SO_REUSEADDR");
      free (rtcp_recv);
      return -1;
    }
    
    if (fcntl(rtcp_recv->soc, F_SETFL, O_NONBLOCK) < 0) {
      perror("add_rtcp_recv: fcntl O_NONBLOCK");
      free (rtcp_recv);
      return -1;
    }

    if (bind (rtcp_recv->soc, (struct sockaddr *)&rtcp_recv->s_addr, get_ss_len(&rtcp_recv->s_addr)) < 0) {
      perror("add_rtcp_recv: bind");
      free (rtcp_recv);
      return (-1);
    }

    TAILQ_INSERT_TAIL(rtcp_recv_list, rtcp_recv, chain);
  }

  return (1);
}

static int
_prepare_rtcp_send_common (struct dest_list_head *dest_list_head, struct rtcp_recv_list_head *rtcp_recv_list)
{
  struct dest_obj *dest;

  for (dest = dest_list_head->tqh_first; dest != NULL;
       dest = (struct dest_obj *)dest->chain.tqe_next) {
#ifdef FF_DVTS
    u_int16_t port = get_ss_port(&dest->rtp_s_addr) + 10;
	printf("Send RTCP socket, port[%d]\n", port);
#else
    u_int16_t port = get_ss_port(&dest->rtp_s_addr) + 1;
#endif
    
	dest->rtcp_s_addr = dest->rtp_s_addr;
    set_ss_port(&dest->rtcp_s_addr, port);

    if (_add_rtcp_recv(rtcp_recv_list, &dest->rtcp_s_addr) < 0) {
      return (-1);
	  
    printf("RTCP dest/port : %s %d\n",
	   dest->dest_str, get_ss_port(&dest->rtcp_s_addr));
    }
  }
    
  return(1);
}

#if 0
static int
_prepare_rtcp_recv_common (struct dest_list_head *dest_list_head)
{
  struct dest_obj *dest;

  for (dest = dest_list_head->tqh_first; dest != NULL;
       dest = (struct dest_obj *)dest->chain.tqe_next){
    dest->rtcprecv.s_addr = dest->rtcp.s_addr;
    dest->rtcprecv.soc = socket(dest->rtcprecv.s_addr.ss_family, SOCK_DGRAM, 0);
    if (dest->rtcprecv.soc < 0) {
      perror("socket");
      return(dest->rtcprecv.soc);
    }
    
    if(dest->rtcprecv.s_addr.ss_family == AF_INET){
      ((struct sockaddr_in *)&dest->rtcprecv.s_addr)->sin_addr.s_addr = htonl(INADDR_ANY);
    } else if(dest->rtcprecv.s_addr.ss_family == AF_INET6){
      ((struct sockaddr_in6 *)&dest->rtcprecv.s_addr)->sin6_addr = in6addr_any;
    }
    
    if (bind (dest->rtcprecv.soc, (struct sockaddr *)&dest->rtcprecv.s_addr, 
	      get_ss_len(&dest->rtcprecv.s_addr)) < 0){
      perror("rtcp recv bind");
      return(-1);
    }
  }
  return (1);
}
#endif

static int
_prepare_rtcp_send (struct hdvsend_param *hdvsend_param)
{
  printf("RTCP DESTINATIONS with AUDIO & VIDEO\n");
  return(_prepare_rtcp_send_common(&hdvsend_param->dest_list, &hdvsend_param->rtcp_recv_list));
}

static int
_prepare_rtcp_audio_send (struct hdvsend_param *hdvsend_param)
{
  printf("RTCP DESTINATIONS with AUDIO ONLY\n");
  return(_prepare_rtcp_send_common(&hdvsend_param->dest_audio_list, &hdvsend_param->rtcp_recv_list));
}

#if 0
static int
_prepare_rtcp_recv (struct hdvsend_param *hdvsend_param)
{
  struct dest_obj *dest = NULL;

  for (dest = hdvsend_param->dest_list; dest != NULL; dest = dest->chain.tqe_next) {
    
  }

  for (dest = hdvsend_param->dest_audio_list; dest != NULL; dest = dest->chain.tqe_next) {
    
  }

  return(_prepare_rtcp_recv_common(&hdvsend_param->dest_list_head));
}

static int
_prepare_rtcp_audio_recv (struct hdvsend_param *hdvsend_param)
{
  printf("RTCP DESTINATIONS with AUDIO ONLY\n");
  return(_prepare_rtcp_recv_common(&hdvsend_param->audio_dest_list_head));
}
#endif

static int
_prepare_send_sockaddr(char *dsthost, char *serv,
		       struct sockaddr_storage *saddr)
{
  struct addrinfo hints, *res;
  int gai_ret;

  memset(&hints, 0, sizeof(hints));
  hints.ai_socktype = SOCK_DGRAM;
  hints.ai_protocol = IPPROTO_UDP;
  if (saddr->ss_family != 0) {
    hints.ai_family = saddr->ss_family;
  }

  gai_ret = getaddrinfo(dsthost, serv, &hints, &res);
  if (gai_ret != 0) {
    fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(gai_ret));
    printf("%s\n", dsthost);
    return(-1);
  }

  memcpy(saddr, res->ai_addr, res->ai_addrlen);
  freeaddrinfo(res);
  return(1);
}
